#ifdef TSOL
/*
 * Copyright (c) 1996,, 1997, 1998, 1999 by Sun Microsystems, Inc.
 * All rights reserved.
 */

#pragma ident	"@(#)tsol_rpcb.c	7.4	99/12/14 SMI; TSOL 2.x"

/*
 *   These are the functions called only by the
 *   Trusted Solaris version of rpcbind.  They
 *   are called by hooks in the main body of the
 *   rpcbind code.
 */

#include "rpcbind.h"

#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <dirent.h>
#include <syslog.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/tsol/tndb.h>
#include <tsol/priv.h>
#include <tsol/label.h>
#include <tsol/auth_names.h>
#include <tsol/pattr.h>
#include <tsix/t6attrs.h>
#include <rpcsvc/bootparam_prot.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <bsm/audit.h>
#include <sys/param.h>
#include <bsm/libbsm.h>
#include <stdarg.h>

/* UTILITIES */
static int tsol_trusted_path(void);
static void tsol_setprivs(void);
static void checkprivs(void);
static void print_privs(priv_set_t privs);
static void tsol_priv_on(int count, ...);
static void tsol_priv_off(int count, ...);

/* Clone Client Attrs Utilities */

static int clone_client_attrs(t6attr_t attrblk, t6mask_t mask);
static char *chk(char *p);
static void * setup(int attr_type, t6attr_t attrblk, t6mask_t mask,
    priv_t priv);

#define	NAME_ARRAY_SIZE 4000 /* chars in array for priv names */

/*
 * ************************************************************
 *
 * RPCBIND STARTUP
 *
 * ************************************************************
 */

/*
 *  tsol_rpcbind_start_check
 *
 *   This first checks that the process has all privileges in
 *   the permitted privileges set, that it has
 *   the trusted path attribute, and that it is running
 *   with an SL of ADMIN_HIGH.  Then it sets the process
 *   clearance to ADMIN_HIGH so SLs on child processes
 *   used for the RPCB_CALLIT function can be set to
 *   whatever level is needed.  If all privileges are not
 *   in the permitted set, a warning message is printed.
 *   If the SL is not ADMIN_HIGH, it is reset to ADMIN_HIGH.
 *   If rpcbind is not run from the trusted path, an error
 *   message is printed and an exit is performed.
 *
 */
extern void
tsol_rpcbind_start_check()
{
	bclear_t clearance;
	bclabel_t plabel;
	bslabel_t admin_high_sl;

	if (debugging)
		fprintf(stderr, "tsol_rpcbind_start_check: called\n");

	if (!tsol_trusted_path()) {
		fprintf(stderr, "rpcbind: not called from trusted path\n");
		exit(1);
	}
	if (setpattr(PAF_DISKLESS_BOOT, 1) != 0) {
		fprintf(stderr, "rpcbind: can't set PAF_DISKLESS_BOOT\n");
		exit(1);
	}

	/*
	 * Check for required privileges (those used by rpcbind) and
	 * expected privileges (all privileges are expected).  If
	 * required privileges are missing an error message is printed
	 * and an exit is performed.  If expected privileges are missing
	 * a warning message is printed and execution continues.
	 */

	checkprivs();

	/*
	 * set the clearance to ADMIN_HIGH, so that we can set
	 * process SLs as high as needed.
	 */

	tsol_priv_on(1, PRIV_PROC_SETCLR);
	bclearhigh(&clearance);
	if (setclearance(&clearance) != 0) {
		(void) syslog(LOG_ERR,
		    "rpcbind: Cannot set clearance to ADMIN_HIGH");
		exit(1);
	}
	tsol_priv_off(1, PRIV_PROC_SETCLR);

	/* Check that our SL is ADMIN_HIGH, set it if not */

	if (getcmwplabel(&plabel) != 0)
		fprintf(stderr, "rpcbind: Cannot get process label");
	else {
		bslhigh(&admin_high_sl);
		if (!blequal(&admin_high_sl, bcltosl(&plabel))) {
			fprintf(stderr,
			    "rpcbind: resetting SL to ADMIN_HIGH\n");
			tsol_priv_on(1, PRIV_PROC_SETSL);
			setcsl(&plabel, &admin_high_sl);
			if (setcmwplabel(&plabel, SETCL_SL) != 0) {
				fprintf(stderr,
				    "Cannot set SL to ADMIN_HIGH\n");
				exit(1);
			}
			tsol_priv_off(1, PRIV_PROC_SETSL);
		}
	}

	/*
	 * Turn off all effective privileges except:
	 *	proc_nofloat
	 *	net_downgrade_sl
	 *	net_upgrade_sl
	 */

	tsol_setprivs();
}

/*
 * tsol_trusted_path
 *
 *	Parameters:
 *		none.
 *
 *	Returns:
 *		1 -> trusted path attribute is on
 *		0 -> trusted path attribute is off
 *			or can not be checked
 */
static int
tsol_trusted_path()
{
	pattr_flag_t sh_attrs;

	if (debugging)
		fprintf(stderr, "rpcbind: checking for trusted path\n");
	/*
	 *  check for trusted path attribute
	 *	return 1 if set, 0 if not.
	 */
	if (getpattr(PAF_TRUSTED_PATH, &sh_attrs) != 0) {
		fprintf(stderr, "rpcbind: could not check process"
			" trusted path attribute: %s\n", strerror(errno));
		return (0);
	} else if (sh_attrs == 0)
		return (0);
	else
		return (1);
}

/*
 * tsol_setprivs -- Turn off all effective and inheritable privileges
 *                 except proc_nofloat
 *
 */
static void
tsol_setprivs()
{
	/*
	 * Turn off all effective privileges.
	 */
	if (__init_effective_priv() == -1) {
		fprintf(stderr,
			"rpcbind: error setting effective privileges"
			" : %s\n", strerror(errno));
		exit(1);
	}

	/* Turn off all inheritable privileges */
	if (__init_inheritable_priv() == -1) {
		fprintf(stderr,
			"rpcbind: error setting inheritable privileges"
			" : %s\n", strerror(errno));
		exit(1);
	}

	/*
	 * Set net_downgrade_sl and net_upgrade_sl for duration of
	 * process.  rpcbind must set the SL of reply messages and
	 * error returns to that of the requestor.  This is done in
	 * numerous locations throughout rpcbind, so the two privileges
	 * are left on to reduce the number of coding changes to the
	 * base code.
	 */
	tsol_priv_on(2, PRIV_NET_DOWNGRADE_SL, PRIV_NET_UPGRADE_SL);

}

/*
 * checkprivs -- Check that we have all privileges
 *
 */
static void
checkprivs()
{
	priv_set_t required_privs;
	priv_set_t expected_privs;
	priv_set_t permitted_privs;
	priv_set_t unpermitted_privs;
	char privbuf[1024];
	int privbuflen;

	if (debugging)
		fprintf(stderr, "rpcbind: checking for needed privileges\n");

	/* get process's permitted privileges */
	if (getppriv(PRIV_PERMITTED, &permitted_privs) == -1) {
		fprintf(stderr, "rpcbind: error getting privilege set"
			" : %s\n", strerror(errno));
		exit(1);
	}

	if (debugging)
		print_privs(permitted_privs);

	PRIV_EMPTY(&required_privs);
	PRIV_ASSERT(&required_privs, PRIV_PROC_SETCLR);
	PRIV_ASSERT(&required_privs, PRIV_PROC_SETSL);
	PRIV_ASSERT(&required_privs, PRIV_PROC_SETID);
	PRIV_ASSERT(&required_privs, PRIV_NET_DOWNGRADE_SL);
	PRIV_ASSERT(&required_privs, PRIV_NET_UPGRADE_SL);
	PRIV_ASSERT(&required_privs, PRIV_NET_MAC_READ);
	PRIV_ASSERT(&required_privs, PRIV_NET_PRIVADDR);
	PRIV_ASSERT(&required_privs, PRIV_SYS_AUDIT);

	if (!PRIV_ISSUBSET(&required_privs, &permitted_privs)) {
		unpermitted_privs = permitted_privs;
		PRIV_INVERSE(&unpermitted_privs);
		PRIV_INTERSECT(&required_privs, &unpermitted_privs);
		privbuflen = sizeof (privbuf);
		priv_set_to_str(&unpermitted_privs, ',',
		    privbuf, &privbuflen);
		fprintf(stderr, "rpcbind: required privileges missing: %s\n",
			privbuf);
		exit(1);
	}

	PRIV_FILL(&expected_privs);
	if (!PRIV_ISSUBSET(&expected_privs, &permitted_privs)) {
		fprintf(stderr,
		    "rpcbind: WARNING - should be run with all privileges\n");
	}
}

/*
 * ************************************************************
 *
 * RPCBIND NEW PROCEDURES
 *
 * ************************************************************
 */

/*
 *  tsol_rpcbproc_dump
 *
 *   Since the mappings maintained by rpcbind now contain SL and
 *   the MLM flag, they can not be returned directly by the DUMP
 *   procedure.  This function goes through the list of mappings
 *   and builds the list to be returned to the requestor.  It
 *   returns only those mappings that are MLMs or are at the
 *   requestor's SL.  The mappings returned do not include the
 *   SL and the MLM flag.
 */
/* ARGSUSED */
char **
tsol_rpcbproc_dump(regp, rqstp, transp)
	RPCB *regp;
	struct svc_req *rqstp;	/* Not used here */
	SVCXPRT *transp;
{
	tsol_rpcblist_ptr mappingp;
	rpcblist_ptr result, rbl, last;
	static char **resultp;
	RPCB *a;
	bslabel_t *slp;
	int mcount;

	result = NULL;
	last = NULL;

#ifdef RPCBIND_DEBUG
	fprintf(stderr, "rpcbind: TSOL_RPCBPROC_DUMP:\n");
#endif
	/* Get the SL of the requestor */
	slp = (bslabel_t *)t6get_attr(T6_SL, transp->xp_tsol_incoming_attrsp);
	if (slp == NULL) {
		fprintf(stderr, "rpcbind: t6get_attr error getting SL\n");
		return (NULL);
	}

	mcount = 0;
	for (mappingp = list_rbl;
		mappingp != NULL;
		mappingp = mappingp->rpcb_next) {

		if (!blequal(slp, &mappingp->rpcb_map.r_tsol_senslabel) &&
	    ((mappingp->rpcb_map.r_tsol_flags & TSOL_MULTILEVEL_FLAG) == 0))
			continue;

		rbl = (rpcblist_ptr) malloc((u_int)sizeof (RPCBLIST));
		if (rbl == NULL) {
			fprintf(stderr, "rpcbind: malloc error\n");
			tsol_free_rpcblist((char *)result);
			return (NULL);
		}
		++mcount;
		a = &(rbl->rpcb_map);
		a->r_prog = mappingp->rpcb_map.r_prog;
		a->r_vers = mappingp->rpcb_map.r_vers;
		a->r_netid = mappingp->rpcb_map.r_netid;
		a->r_addr = mappingp->rpcb_map.r_addr;
		a->r_owner = mappingp->rpcb_map.r_owner;
		rbl->rpcb_next = NULL;

		if (result == NULL)
			result = rbl;
		else
			last->rpcb_next = rbl;
		last = rbl;

	}
	resultp = (char **)&result;
#ifdef RPCBIND_DEBUG
	fprintf(stderr,
	    "TSOL rpcbind: malloc %d entries plus pointer\n", mcount);
#endif
	return (resultp);
}

/*
 *  tsol_rpcbproc_dumpall
 *
 *   This function supports the new DUMPALL procedure.  It builds the
 *   list of mappings to be returned to the requestor.  If the requestor
 *   has net_mac_read privilege, all mappings are returned.  If not
 *   all MLMs and all mappings dominated by the requestor's SL are
 *   returned.  The mappings returned include the SL and MLM flag.
 */
/* ARGSUSED */
char **
tsol_rpcbproc_dumpall(regp, rqstp, transp)
	RPCB *regp;
	struct svc_req *rqstp;	/* Not used here */
	SVCXPRT *transp;
{
	tsol_rpcblist_ptr mappingp;
	tsol_rpcblist_ptr result, rbl, last;
	static char **resultp;
	TSOL_RPCB *a;
	bslabel_t *slp;
	priv_set_t *privp;
	int mcount;

	result = NULL;
	last = NULL;

#ifdef RPCBIND_DEBUG
	fprintf(stderr, "rpcbind: TSOL_RPCBPROC_DUMPALL:\n");
#endif
	/* Get the privileges of the requestor */
	privp = (priv_set_t *)t6get_attr(T6_PRIVILEGES,
	    transp->xp_tsol_incoming_attrsp);
	if (privp == NULL) {
		fprintf(stderr,
		    "rpcbind: t6get_attr error getting privileges\n");
		return (NULL);
	}

	/* Get the SL of the requestor */
	slp = (bslabel_t *)t6get_attr(T6_SL, transp->xp_tsol_incoming_attrsp);
	if (slp == NULL) {
		fprintf(stderr, "rpcbind: t6get_attr error getting SL\n");
		return (NULL);
	}

	mcount = 0;
	for (mappingp = list_rbl;
		mappingp != NULL;
		mappingp = mappingp->rpcb_next) {

		if (!bldominates(slp, &mappingp->rpcb_map.r_tsol_senslabel) &&
	    ((mappingp->rpcb_map.r_tsol_flags & TSOL_MULTILEVEL_FLAG) == 0) &&
		    (!PRIV_ISASSERT(privp, PRIV_NET_MAC_READ)))
			continue;

		rbl = (tsol_rpcblist_ptr)malloc((u_int)sizeof (TSOL_RPCBLIST));
		if (rbl == NULL) {
			fprintf(stderr, "rpcbind: malloc error\n");
			tsol_free_tsol_rpcblist((char *)result);
			return (NULL);
		}
		++mcount;
		a = &(rbl->rpcb_map);
		a->r_prog = mappingp->rpcb_map.r_prog;
		a->r_vers = mappingp->rpcb_map.r_vers;
		a->r_netid = mappingp->rpcb_map.r_netid;
		a->r_addr = mappingp->rpcb_map.r_addr;
		a->r_owner = mappingp->rpcb_map.r_owner;
		a->r_tsol_senslabel = mappingp->rpcb_map.r_tsol_senslabel;
		a->r_tsol_flags = mappingp->rpcb_map.r_tsol_flags;
		rbl->rpcb_next = NULL;

		if (result == NULL)
			result = rbl;
		else
			last->rpcb_next = rbl;
		last = rbl;

	}
	resultp = (char **)&result;
#ifdef RPCBIND_DEBUG
	fprintf(stderr,
	    "TSOL rpcbind: malloc %d entries plus pointer\n", mcount);
#endif
	return (resultp);
}
#ifdef PORTMAP
/*
 *  tsol_pmaplist_dump
 *
 *    This function peforms the same processing as tsol_rpcbproc_dump except
 *    that it processes the portmap mappings list.
 */
/* ARGSUSED */
char **
tsol_pmaplist_dump(rqstp, transp)
	struct svc_req *rqstp;
	SVCXPRT *transp;
{
	TSOL_PMAPLIST *mappingp;
	pmaplist_ptr result, pml, last;
	static char **resultp;
	PMAP *a;
	bslabel_t *slp;
	int mcount;

	result = NULL;
	last = NULL;

#ifdef RPCBIND_DEBUG
	fprintf(stderr, "rpcbind: TSOL_PMAPLIST_DUMP:\n");
#endif
	/* Get the SL of the requestor */
	slp = (bslabel_t *)t6get_attr(T6_SL, transp->xp_tsol_incoming_attrsp);
	if (slp == NULL) {
		fprintf(stderr, "rpcbind: t6get_attr error getting SL\n");
		return (NULL);
	}

	mcount = 0;
	for (mappingp = list_pml;
		mappingp != NULL;
		mappingp = mappingp->pml_next) {

		if (!blequal(slp, &mappingp->pml_map.pm_tsol_senslabel) &&
	    ((mappingp->pml_map.pm_tsol_flags & TSOL_MULTILEVEL_FLAG) == 0))
			continue;

		pml = (pmaplist_ptr)
		    malloc((u_int)sizeof (PMAPLIST));
		if (pml == NULL) {
			fprintf(stderr, "rpcbind: malloc error\n");
			tsol_free_pmaplist((char *)result);
			return (NULL);
		}
		++mcount;
		a = &(pml->pml_map);
		a->pm_prog = mappingp->pml_map.pm_prog;
		a->pm_vers = mappingp->pml_map.pm_vers;
		a->pm_port = mappingp->pml_map.pm_port;
		a->pm_prot = mappingp->pml_map.pm_prot;
		pml->pml_next = NULL;

		if (result == NULL)
			result = pml;
		else
			last->pml_next = pml;
		last = pml;

	}
	resultp = (char **)&result;
#ifdef RPCBIND_DEBUG
	fprintf(stderr,
	    "TSOL rpcbind: malloc %d entries plus pointer\n", mcount);
#endif
	return (resultp);
}
#endif /* PORTMAP */

/*
 * Since tsol_rpcbproc_dump, tsol_rpcbproc_dumpall, and tsol_pmaplist_dump
 * malloc storage for their results, the following three functions exist to
 * free the results.
 */

void
tsol_free_rpcblist(char *result)
{
	rpcblist_ptr mappingp, rbl;
	int mcount;

	if (result == NULL)
		return;

	mcount = 0;
	mappingp = *(rpcblist_ptr *)result;
	while (mappingp != NULL) {

		rbl = mappingp;
		mappingp = mappingp->rpcb_next;
		free(rbl);
		++mcount;
	}
#ifdef RPCBIND_DEBUG
	fprintf(stderr,
	    "TSOL rpcbind: freeing %d entries plus pointer\n", mcount);
#endif
}

#ifdef PORTMAP
void
tsol_free_pmaplist(char *result)
{
	pmaplist_ptr mappingp, pml;
	int mcount;

	if (result == NULL)
		return;

	mcount = 0;
	mappingp = *(pmaplist_ptr *)result;
	while (mappingp != NULL) {

		pml = mappingp;
		mappingp = mappingp->pml_next;
		free(pml);
		++mcount;
	}
#ifdef RPCBIND_DEBUG
	fprintf(stderr,
	    "TSOL rpcbind: freeing %d entries plus pointer\n", mcount);
#endif
}
#endif /* PORTMAP */

void
tsol_free_tsol_rpcblist(char *result)
{
	tsol_rpcblist_ptr mappingp, rbl;
	int mcount;

	if (result == NULL)
		return;

	mcount = 0;
	mappingp = *(tsol_rpcblist_ptr *)result;
	while (mappingp != NULL) {

		rbl = mappingp;
		mappingp = mappingp->rpcb_next;
		free(rbl);
		++mcount;
	}
#ifdef RPCBIND_DEBUG
	fprintf(stderr,
	    "TSOL rpcbind: freeing %d entries plus pointer\n", mcount);
#endif
}

/*
 * This function allocates t6 opaque structures and sets pointers
 * to them in the SVCXPRT handle so the rpc library can store the
 * attributes of the requestor.
 */

void
tsol_alloc_attrs(SVCXPRT *handlep)
{
	/* we don't need T6M_AUDIT_ID, T6M_AUDIT_INFO and T6M_GROUPS */
	handlep->xp_tsol_incoming_attrsp = t6alloc_blk(T6M_ALL_ATTRS &
	    ~(T6M_AUDIT_ID | T6M_AUDIT_INFO | T6M_GROUPS));
	if (handlep->xp_tsol_incoming_attrsp == NULL) {
		fprintf(stderr,
		    "rpcbind: Can't create incoming attr buffer\n");
		exit(1);
	}
	handlep->xp_tsol_outgoing_attrsp = t6alloc_blk(T6M_SL);
	if (handlep->xp_tsol_outgoing_attrsp == NULL) {
		fprintf(stderr,
		    "rpcbind: Can't create outgoing attr buffer\n");
		exit(1);
	}
	return;

}

/*
 * When rpcbind sends replies to all requests except CALLIT, it
 * sets the SL of the reply to the SL of the requestor.
 */

int
tsol_svc_set_reply_sl(SVCXPRT *xprt)
{
	bslabel_t *slp;
#ifdef RPCBIND_DEBUG
	char *label;
	fprintf(stderr, "rpcbind: ===> setting reply SL - ");
#endif
	if (xprt->xp_tsol_incoming_attrsp == NULL)
		return (-1);

	/* Get the SL of the requestor */
	slp = (bslabel_t *)t6get_attr(T6_SL, xprt->xp_tsol_incoming_attrsp);
	if (slp == NULL) {
		fprintf(stderr,
		    "rpcbind: Can't get requestor's SL\n");
		return (-1);
	}
#ifdef RPCBIND_DEBUG
	label = sbsltos(slp, 0);
	fprintf(stderr, "to %s - ",
	    label == NULL ? "*NULL*" : label);
#endif
	/* Set the SL of the response */

	if (t6set_attr(T6_SL, slp, xprt->xp_tsol_outgoing_attrsp) < 0) {
		fprintf(stderr,
		    "rpcbind: Can't set SL of response\n");
		return (-1);
	}
#ifdef RPCBIND_DEBUG
	fprintf(stderr, "done\n");
#endif
	return (0);
}

/*
 * At various times, tests must be performed with the SL and privileges
 * that came in a request to rpcbind.  This function retrieves the SL
 * and privileges that came in the last request associated with a
 * specific SVCXPRT handle.
 */

int
tsol_get_incoming_attrs(SVCXPRT *transp, bslabel_t **slpp, priv_set_t **prvpp)
{
#ifdef RPCBIND_DEBUG
	char *label;
	int *pidp;
	int pid;
#endif
	*slpp = (bslabel_t *)t6get_attr(T6_SL,
	    transp->xp_tsol_incoming_attrsp);
	if (*slpp == NULL) {
		fprintf(stderr, "rpcbind: t6get_attr error getting SL\n");
		return (-1);
	}
#ifdef RPCBIND_DEBUG
	pidp = (int *)t6get_attr(T6_PID, transp->xp_tsol_incoming_attrsp);
	pid = *pidp;
	label = sbsltos(*slpp, 0);
	fprintf(stderr, "\n - incoming SL of %s from pid %d - \n",
	    label == NULL ? "*NULL*" : label, pid);
#endif
	*prvpp = (priv_set_t *)t6get_attr(T6_PRIVILEGES,
	    transp->xp_tsol_incoming_attrsp);
	if (*prvpp == NULL) {
		fprintf(stderr,
		    "rpcbind: t6get_attr error getting privileges\n");
		return (-1);
	}
	return (0);
}
/*
 * When the CALLIT procedure is processed, a child process is forked to make
 * the call to the server on behalf of the client.  Before the request is
 * sent to the server, the attributes received in the request from the client
 * are used to set the process attributes of the child process.  The child
 * process is used because some attributes are not supported by
 * t6set_endpt_default for the TSOL network host type.
 *
 * This function handles the creation of the child and waits for it
 * to exit before the parent process continues execution.
 */

/*
 * ALERT - When new attributes are handled by libt6 they
 * should be added here
 */

static t6mask_t clone_mask = T6M_SL + T6M_CLEARANCE + T6M_IL + T6M_PID +
	T6M_UID + T6M_GID + T6M_GROUPS + T6M_PRIVILEGES;

int
tsol_t_sndudata(int fd,	struct t_unitdata *tu_data, SVCXPRT *transp)
{
	int stat;
	int pid;
	int wait_pid;
	int child_status;
	t6mask_t saved_mask;

	/* Fork a child process to send the remote call */

	pid = fork();
	if (pid > 0) {

		/* Parent process waits for child to exit, then returns */

		if (debugging)
			fprintf(stderr,
			    "Remote call child (%d) created by (%d)\n",
			    pid, getpid());

		wait_pid = waitpid(pid, &child_status, 0);

		if (debugging)
			fprintf(stderr, "Remote call child (%d) status %d\n",
			    wait_pid, child_status);

		return (child_status);

	} else if (pid < 0) {

		/* Fork failed */
		fprintf(stderr, "rpcbind: fork failed\n");
		return (-1);

	} else {
		/* Child process */

		if (debugging)
			fprintf(stderr,
			    "Remote call child (%d) running\n", getpid());

		if (debugging) {
			bslabel_t *slp;
			priv_set_t *privp;
			tsol_get_incoming_attrs(transp, &slp, &privp);
		}

		/* Set process attributes to those of requestor */

		if (clone_client_attrs(transp->xp_tsol_incoming_attrsp,
		    clone_mask) < 0) {
			fprintf(stderr,
			    "rpcbind: attribute cloning failed\n");
			exit(-1);
		}

		/* Send the data */

		if (debugging)
			fprintf(stderr,
			    "sending %d bytes\n", tu_data->udata.len);

		/*
		 * net_reply_equal is forced off because it will
		 * interfere with the setting of the SL on the
		 * packet being sent.
		 */
		tsol_priv_off(1, PRIV_NET_REPLY_EQUAL);

		stat = t_sndudata(fd, tu_data);

		/* All done, so child process exits */

		if (debugging)
			fprintf(stderr,
			    "Remote call child (%d) exiting\n", getpid());

		exit(stat);
	}
	/* NOTREACHED */
}

/*
 * While processing the CALLIT procedure,
 * another child process is forked to receive the reply from the server and
 * forward it to the client.  In that child, the attributes received in the
 * reply from the server are used to set the attributes of the child process.
 *
 * This function handles the creation of the child and waits for it
 * to exit before the parent process continues execution.
 */

void
tsol_svc_sendreply(SVCXPRT *transp, xdrproc_t xdr_rmtcall_result, char *a)
{
	int pid;
	int wait_pid;
	int child_status;
	t6attr_t save_attrp;
	t6mask_t saved_mask;

	/* Fork a child process to send the callit reply */

	pid = fork();
	if (pid > 0) {

		/* Parent process waits for child to exit, then returns */

		if (debugging)
			fprintf(stderr,
			    "Callit reply child (%d) created by (%d)\n",
			    pid, getpid());

		wait_pid = waitpid(pid, &child_status, 0);

		if (debugging)
			fprintf(stderr, "Callit reply child (%d) status %d\n",
			    wait_pid, child_status);

		return;
	} else if (pid < 0) {

		/* Fork failed */
		fprintf(stderr, "rpcbind: fork failed\n");
		return;
	} else {
		/* Child process */

		if (debugging)
			fprintf(stderr,
			    "Callit reply child (%d) running\n", getpid());

		if (debugging) {
			bslabel_t *slp;
			priv_set_t *privp;
			tsol_get_incoming_attrs(transp, &slp, &privp);
		}

		/* Set process attributes to those of requestor */

		if (clone_client_attrs(transp->xp_tsol_incoming_attrsp,
		    clone_mask) < 0) {
			fprintf(stderr, "attribute cloning failed\n");
			exit(-1);
		}

		/* Send the reply */

		save_attrp = transp->xp_tsol_outgoing_attrsp;
		transp->xp_tsol_outgoing_attrsp = NULL;

		/*
		 * net_reply_equal is forced off because it will
		 * interfere with the setting of the SL on the
		 * packet being sent.
		 */
		tsol_priv_off(1, PRIV_NET_REPLY_EQUAL);

		svc_sendreply(transp, (xdrproc_t)xdr_rmtcall_result, (char *)a);

		transp->xp_tsol_outgoing_attrsp = save_attrp;

		/* All done, so child process exits */

		if (debugging)
			fprintf(stderr,
			    "Callit reply child (%d) exiting\n", getpid());
		exit(0);
	}
}
/*
 * When rpcbind binds to a port, it needs to be an MLP.  The calls to t_bind
 * in rpcbind have been replaced with calls to this function.
 */

int
tsol_t_bind(int fd, register struct t_bind *req, register struct t_bind *ret)
{
	int stat;

	/*
	 * Make the bind a Multi-level Port (MLP)
	 */
	tsol_priv_on(2, PRIV_NET_MAC_READ, PRIV_NET_PRIVADDR);

	stat = t_bind(fd, req, ret);

	tsol_priv_off(2, PRIV_NET_MAC_READ, PRIV_NET_PRIVADDR);

	if (debugging)
		fprintf(stderr, "Binding fd (%d) - status (%d)\n", fd, stat);

	return (stat);
}

/*
 * The check_bound function had to be modified for TSOL to do the bind
 * check based on the MLM flag and SL for the mapping being checked.
 * The following two functions are used to set the net_mac_read privilege
 * and SL of the endpt properly before and after the bind check.
 */

int
tsol_set_bind_state(int mlm_flag, bslabel_t *slp, int out_fd)
{
	static t6attr_t endpt_attrsp = NULL;

	if (debugging)
		fprintf(stderr,
		    " - setting state before check_bound fd %d\n", out_fd);

	if ((mlm_flag & TSOL_MULTILEVEL_FLAG) != 0) {

		/* If it is an MLM, then just set net_mac_read privilege */

		tsol_priv_on(1, PRIV_NET_MAC_READ);

	} else {

		/* It is not an MLM, so set the SL on the end point */

		if (endpt_attrsp == NULL) {
			endpt_attrsp = t6alloc_blk(T6M_SL);
			if (endpt_attrsp == NULL) {
				fprintf(stderr,
				    "rpcbind: t6alloc_blk failed\n");
				return (0);
			}
		}

		/* Set endpt mask to set just the SL */

		if (t6set_endpt_mask(out_fd, T6M_SL) < 0) {
			fprintf(stderr,
			    "rpcbind: t6set_endpt_mask failed\n");
			return (0);
		}

		/* Set the SL */

		if (t6set_attr(T6_SL, slp, endpt_attrsp) < 0) {
			fprintf(stderr,
			    "rpcbind: Can't set SL of response\n");
			return (0);
		}

		if (t6set_endpt_default(out_fd, T6M_SL, endpt_attrsp) < 0) {
			fprintf(stderr,
			    "rpcbind: t6set_endpt_default failed\n");
			return (0);
		}
	}
	return (1);
}

/* ARGSUSED */
void
tsol_reset_bind_state(int mlm_flag, bslabel_t *slp, int out_fd)
{
	if (debugging)
		fprintf(stderr,
		    " - reset state after check_bound fd %d\n", out_fd);

	if ((mlm_flag & TSOL_MULTILEVEL_FLAG) != 0) {

		/* If it is an MLM, then just clear net_mac_read privilege */

		tsol_priv_off(1, PRIV_NET_MAC_READ);

	} else {

		/* It is not an MLM, so reset the mask on the end point */

		if (debugging)
			fprintf(stderr,
			    " -- reset endpt mask to zero\n");
		if (t6set_endpt_mask(out_fd, 0) < 0) {
			fprintf(stderr,
			    "rpcbind: t6set_endpt_mask failed\n");
			return;
		}
	}
}
/*
 * All the calls to set or unset a mapping must be checked to see if
 * they are for a privileged port (tcp and udp only).  If they are
 * requests for a privileged port, they are only allowed if PRIV_NET_PRIVADDR
 * privilege is asserted in the requestor's privileges.
 *
 * This function is called to make the check.  It returns false if the
 * port is privileged and the caller does not have PRIV_NET_PRIVADDR.  It
 * also returns false on any error condition.  It returns true otherwise.
 */

int
tsol_privport_chk(char *netid, char *uaddr, priv_set_t *privp)
{
	int h1, h2, h3, h4, p1, p2, port;

	if ((strcmp(netid, udptrans) != 0) &&
	    (strcmp(netid, tcptrans) != 0)) {
		/* It is not TCP or UDP so no privilege port check is needed */
		return (1);
	}

	/* interpret the universal address for TCP/IP */
	if (sscanf(uaddr, "%d.%d.%d.%d.%d.%d",
		&h1, &h2, &h3, &h4, &p1, &p2) != 6)
		return (0);

	port = ((p1 & 0xff) << 8) + (p2 & 0xff);

	if (debugging)
		fprintf(stderr, " - privilege port check on port %d\n", port);

	if ((port < IPPORT_RESERVED) &&
	    (!PRIV_ISASSERT(privp, PRIV_NET_PRIVADDR)))
		return (0);

	return (1);
}


/*
 * As mentioned previously,
 *
 * When the CALLIT procedure is processed a child process is forked to make
 * the call to the server on behalf of the client.  Before the request is
 * sent to the server, the attributes received in the request from the client
 * are used to set the process attributes of the child process.
 *
 * Another child process is forked to receive the reply from the server and
 * forward it to the client.  In that child, the attributes received in the
 * reply from the server are used to set the attributes of the child process.
 *
 * This function sets process attributes to match those in a t6attr_t structure.
 * If any error occurs, return -1, leaving process attributes in an
 * indeterminate state.
 */

static int
clone_client_attrs(t6attr_t attrblk, t6mask_t mask)
{
	bclear_t *clearp;
	bclear_t cur_clear;
	bclear_t high_clear;
	bslabel_t *slabelp;
	bslabel_t *cur_sl;
	bclabel_t cur_cmwlabel;
	bclabel_t new_cmwlabel;
	bclear_t new_clearance;
	priv_set_t *privsetp;
	au_id_t *auditidp;
	auditinfo_t *auditinfop;
	uid_t *uidp;
	gid_t *gidp;
	gid_t *groupsp;
	char privbuf[5000];
	int privbuflen;

	if (debugging)
		fprintf(stderr,
		    "- clone_client_attrs: attrblk (%0x) mask (%0x)\n",
		    attrblk, mask);

	if (attrblk == NULL)
		return (0);

	if (getcmwplabel(&cur_cmwlabel) != 0) {
		fprintf(stderr, "%s. Cannot get process label.\n",
		    strerror(errno));
		return (-1);
	}

	if (getclearance(&cur_clear) != 0) {
		fprintf(stderr, "%s. Cannot get process clearance.\n",
		    strerror(errno));
		return (-1);
	}

	new_cmwlabel = cur_cmwlabel;
	cur_sl = bcltosl(&cur_cmwlabel);

	/*
	 * Set SL, IL & CLEARANCE.
	 * The order of doing these must be such that
	 * Clearance >= SL >= IL at all times.
	 */

	slabelp = (bslabel_t *)setup(T6_SL, attrblk, mask, PRIV_PROC_SETSL);
	clearp = (bclear_t *)setup(T6_CLEARANCE, attrblk, mask,
	    PRIV_PROC_SETCLR);

	if (slabelp != NULL)
		setcsl(&new_cmwlabel, slabelp);
	if (clearp == NULL)
		clearp = &cur_clear;

	if (!blequal(slabelp, cur_sl)) {
		/* Keep clearance above SL during transition. */
		if (!blequal(clearp, &cur_clear)) {
			bclearhigh(&high_clear);
			if (setclearance(&high_clear) != 0) {
				(void) syslog(LOG_ERR,
				    "%s. Cannot set temporary "
				    "process clearance to admin_high.",
				    strerror(errno));
				return (-1);
			}
		}

		if (debugging)
			fprintf(stderr,
			    "Setting child process CMW label to %s.\n",
			    chk(sbcltos(&new_cmwlabel, 0)));

		/* Change SL */
		if (setcmwplabel(&new_cmwlabel, SETCL_SL) < 0) {
			(void) syslog(LOG_ERR,
			    "Cannot set process CMW label to %s.",
			    chk(sbcltos(&new_cmwlabel, 0)));
			return (-1);
		}
	}

	/*
	 * ALERT - It appears that when a process with net_reply_equal
	 * sends a network packet, the clearance in the t6 attributes
	 * can be below the SL.  The following code handles this case
	 * and sets the clearance to the SL.
	 */

	new_clearance = *clearp;
	if (!bldominates(&new_clearance, slabelp)) {
		new_clearance = *slabelp;
		setbltype(&new_clearance, SUN_CLR_ID);
	}

	if (!blequal(&new_clearance, &cur_clear)) {
		if (setclearance(&new_clearance) < 0) {
			(void) syslog(LOG_ERR,
			    "Cannot set process clearance to %s.",
			    chk(sbcleartos(&new_clearance, 0)));
			return (-1);
		}
	}

	tsol_priv_off(2, PRIV_PROC_SETSL, PRIV_PROC_SETCLR);

	/* Set UID */
	uidp = (uid_t *)setup(T6_UID, attrblk, mask, PRIV_PROC_SETID);
	if (uidp != NULL) {
		if (seteuid(*uidp) < 0) {
		    fprintf(stderr, "Cannot set process UID to %d.\n", *uidp);
		    return (-1);
		}
		tsol_priv_off(1, PRIV_PROC_SETID);
	}

	/* Set GID */
	gidp = (gid_t *)setup(T6_GID, attrblk, mask, PRIV_PROC_SETID);
	if (gidp != NULL) {
		if (setegid(*gidp) < 0) {
		    fprintf(stderr, "Cannot set process GID to %d.\n", *gidp);
		    return (-1);
		}
		tsol_priv_off(1, PRIV_PROC_SETID);
	}

	/* Set AUDIT_ID */
	auditidp = (au_id_t *)setup(T6_AUDIT_ID, attrblk, mask,
	    PRIV_SYS_AUDIT);
	if (auditidp != NULL) {
		if (setauid(auditidp) < 0) {
		    fprintf(stderr, "Cannot set process audit ID to %d.\n",
			*auditidp);
		    return (-1);
		}
		tsol_priv_off(1, PRIV_SYS_AUDIT);
	}

	/* Skip SESSION_ID */

	/* Set AUDIT_INFO */
	auditinfop = (auditinfo_t *)setup(T6_AUDIT_INFO, attrblk, mask,
	    PRIV_SYS_AUDIT);
	if (auditinfop != NULL) {
		if (setaudit(auditinfop) < 0) {
			fprintf(stderr, "%s. Cannot set process audit info.\n",
			    strerror(errno));
			return (-1);
		}
		tsol_priv_off(1, PRIV_SYS_AUDIT);
	}

	/* Set GROUPS */
	groupsp = (gid_t *)setup(T6_GROUPS, attrblk, mask, PRIV_PROC_SETID);
	if (groupsp != NULL) {
		if (setgroups(groupsp[0], &groupsp[1]) < 0) {
		    fprintf(stderr, "Cannot set process groups.\n");
		    return (-1);
		}
		tsol_priv_off(1, PRIV_PROC_SETID);
	}

	/*
	 * Set PRIVILEGES.
	 * These are set last, after other privilege fiddling.
	 */
	privsetp = (priv_set_t *)setup(T6_PRIVILEGES, attrblk, mask, 0);
	if (privsetp != NULL) {
		if (debugging)
			print_privs(*privsetp);

		if (setppriv(PRIV_SET, PRIV_EFFECTIVE, privsetp) < 0) {
			privbuflen = sizeof (privbuf);
			fprintf(stderr, "Can't set privileges to %s.\n",
			    chk(priv_set_to_str(privsetp, ',', privbuf,
				&privbuflen)));
			return (-1);
		}
	}
	return (0);
}

static char *
chk(char *p)
{
	if (p == NULL)
		return ("<NULL>");
	else return (p);
}

static void *
setup(int attr_type, t6attr_t attrblk, t6mask_t mask, priv_t priv)
{
	void *attrp;

	if ((mask & t6_mask(attr_type)) == 0)
		return (NULL);

	attrp = t6get_attr(attr_type, attrblk);

	if (attrp == NULL) {
		fprintf(stderr,
		    "Error from t6get_attr. Type %d.\n", attr_type);
		return (NULL);
	}

	if (priv != 0)
		tsol_priv_on(1, priv);

	return (attrp);
}

/*
 * ************************************************************
 *
 * MISC UTILITIES
 *
 * ************************************************************
 */

/*
 * print_privs  -- print a privilege set
 *
 *	Parameters:
 *              the privilege set to be printed
 */
static void
print_privs(priv_set_t privs)
{
	static char name_array[NAME_ARRAY_SIZE];
	int n_size;
	char *s;

	n_size = NAME_ARRAY_SIZE;
	s = priv_set_to_str(&privs, ' ', name_array, &n_size);
	fprintf(stderr, "-- %s\n", chk(s));

}

/*
 * tsol_priv_on
 *
 *	Turn on a list of privileges in the effective set.
 *
 *	Parameters:
 *              count - number of privileges to turn on
 *              priv - privileges to turn on
 */
static void
tsol_priv_on(int count, ...)
{
	va_list ap;
	int pnum;
	priv_t priv_id;

	va_start(ap, count);

	for (pnum = 0; pnum < count; pnum++) {

		priv_id = va_arg(ap, priv_t);

		if (set_effective_priv(PRIV_ON, 1, priv_id) != 0)
			fprintf(stderr,
				"rpcbind: cannot set %s privilege (#%d)\n",
				priv_to_str(priv_id), priv_id);
		if (debugging)
			fprintf(stderr, "tsol_priv_on: %s (#%d)\n",
				priv_to_str(priv_id), priv_id);
	}
	va_end(ap);
}
/*
 * tsol_priv_off
 *
 *	Turn off a list of privileges in the effective set.
 *
 *	Parameters:
 *              count - number of privileges to turn off
 *              priv - privileges to turn off
 */
static void
tsol_priv_off(int count, ...)
{
	va_list ap;
	int pnum;
	priv_t priv_id;

	va_start(ap, count);

	for (pnum = 0; pnum < count; pnum++) {

		priv_id = va_arg(ap, priv_t);

		if (set_effective_priv(PRIV_OFF, 1, priv_id) != 0)
			fprintf(stderr,
				"rpcbind: cannot reset %s privilege\n",
				priv_to_str(priv_id));

		if (debugging)
			fprintf(stderr, "tsol_priv_off: %s\n",
				priv_to_str(priv_id));

	}
	va_end(ap);

}

void
tsol_whoami_chstate(SVCXPRT *xprt, u_long prog)
{
	struct sockaddr_in *sin;
	struct sockaddr_in addr;
	struct netbuf nb;

	sin = (struct sockaddr_in *)xprt->xp_rtaddr.buf;

	memcpy(&addr, sin, sizeof (struct sockaddr_in));

	nb.buf = (char *)&addr;
	nb.len = nb.maxlen = sizeof (struct sockaddr_in);

	if (prog == BOOTPARAMPROG && ntohs(sin->sin_port) < IPPORT_RESERVED) {
		tsol_priv_on(1, PRIV_SYS_NET_CONFIG);
		if (chstate(STATE_UNLABELED, &nb) < 0) {
			(void) syslog(LOG_ERR, "chstate failed for %x",
			    sin->sin_addr.s_addr);
		}
		tsol_priv_off(1, PRIV_SYS_NET_CONFIG);
	}
}
#endif /* TSOL */
